#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/shm.h>
#include <time.h>
#include <fcntl.h>
#include <sys/ipc.h>
#include "string.h"
#include "config.h"
#include <signal.h>

pid_t childPID;
pid_t childPIDs[MAX_N];

int* shared_memory;
int shm_id;

void terminateProcess(){ //Function to terminate any child processes
        for (int i = 0; i < MAX_N; i++) {
                if(childPIDs[i] > 1)
                kill(childPIDs[i], SIGTERM);
        }
}

void clean(){ //Function to deallocate shared memory and calls terminate process
terminateProcess();

(shmdt(shared_memory));

if (shmctl(shm_id, IPC_RMID, NULL) == -1) {
        perror("Error deallocating shared memory.");
        exit(1);
        }
        exit(0);
}


void signal_handler(int sig){ //Function to terminate on Ctrl-C
         printf("Terminated with Ctrl-C\n");
        shmdt((void *) shared_memory);
        clean();
        exit(0);
}

void alarm_handler(int arl){ //Function to terminate and clean if timer runs out
        printf("Timer ran out.\n");
        clean();
        exit(0);
}

int main(int argc, char *argv[]) {


signal(SIGINT, signal_handler);

while(1){

int n = 0;
int time = TIMEOUT;

if (argc < 2 || argc > 4){ //Command line arguments to validate user input
    printf("Command line should be formatted like %s -t ss n\n", argv[0]);
    exit(1); 
    }

if (argc >= 2 && strcmp(argv[1], "-t") == 0) {
    time = atoi(argv[2]);
}

if (argc >= 4) {
    n = atoi(argv[3]);
    if (n > MAX_N) {
        printf("n is more than 20, limiting it to 20");
        n = MAX_N;
    }
} else {
    printf("Invalid\n");
    exit(1);
}

key_t key = ftok("shmfile", 65); //Creating key for shared memory
shm_id = shmget(key, sizeof(int) + (MAX_N * sizeof (State)), IPC_CREAT | 0666); //Allocating shared memory

if (shm_id < 0) { //Initalizing flag to idle
    perror("Error allocating shared memory.");
    exit(1);
}

shared_memory = shmat(shm_id, NULL, 0);
int* turn = (int*) shared_memory; //Adding turn to shared memory
State* flag = (State*)(shared_memory + sizeof(int)); //Adding State struc to shared memory

for(int i = 0; i < MAX_N; i++){
        flag[i] = idle;
}

signal(SIGALRM, alarm_handler); //Signal for time

alarm(time);

for(int i = 0; i < n; i++){ //Loop for child processes forking up to user input
        pid_t childPID = fork();

        if (childPID == 0){
                char pid_logical[20];
                sprintf(pid_logical, "%d", i);
                execl("./slave", "./slave", pid_logical, NULL);  //Exec funtion to run slave and pass in child processes
                perror("exec failed");
                exit(1);
        } else if (childPID < 0){
                perror("fork did not work");
                exit(1);
        } else {
               childPIDs[i] = childPID;
        }
}
for (int i = 0; i < n; i++){
        wait(NULL);
}

clean(); 

 return 0;

}
}